# -*- coding: utf-8 -*-
"""
Created on Fri Jun  5 13:19:57 2020

@author: iver_

This program will check how the implemented system behaves with different inputs. This program is mostly for system understanding.
"""


import numpy as np
import matplotlib.pyplot as plt
import numpy.linalg as la
from control import *
from casadi import *
from fp_lib import *
import scipy.io as io

class sim_model(SimulationModel):
    
    def __init__(self, rhs, x0):
        
        self.rhs = rhs
        
        self.x = x0
        
        
    def model_output(self, u):
        y = f(self.x, u)
        self.x = self.x + y*T
        if self.x[2] < 0:
            self.x[2] = 1e-9
        #print(f)
        return self.x
        
g = 9.81
Cc = 2*10**(-5)
A1 = 0.008107
A2 = 0.008107
D1 = 0.1016
D2 = 0.1016
h1 = 200
h2 = 800
hw = 1000
L1 = 500
L2 = 1200
V1 = 4.054
V2 = 9.729
#ESP DATA
B1 = 1.5*10**9
B2 = 1.5*10**9
M = 1.992*10**8
rho = 950
pr = 1.26 * 10 ** 7
f0 = 60 # Hz
#f = [57, 57, 54, 64, 64, 49, 49, 51, 53, 54]
#f = 53 # Hz
#UNKNOWN
PI = 2.32*10**(-9)
#KANSKJE
pm = 20e5
my = 0.025 # Pa*s
CQ = 1 - 2.6266*my + 6.0032*my**2 - 6.8104*my**3 + 2.7944*my**4
CH = 1 - 0.03*my



x = MX.sym('x',3)
p = MX.sym('p',2)
n_states = x.size1()
n_controls = p.size1()

# rhs is the right hand side of the equation
rhs = vertcat((B1/V1)*(PI*(pr-x[0])-x[2]),(B2/V2)*(x[2]-Cc*(sqrt(fmax(10**-5, x[1]-pm))*p[0])),(1/M)*(x[0]-x[1]-rho*g*hw - 0.158*((rho*L1*x[2]**2)/(D1*A1**2))*(my/(rho*D1*x[2]))**(1/4) - 0.158*((rho*L2*x[2]**2)/(D2*A2**2))*(my/(rho*D2*x[2]))**(1/4) + rho*g* (CH*(9.5970e2 + 7.4959e3*((x[2]/CQ)*(f0/p[1])) - 1.2454e6*((x[2]/CQ)*(f0/p[1]))**2)*(p[1]/f0)**2)))

# CasADi function, will map symbolic states and parameters to rhs.
f = Function('f', [x,p], [rhs])

X = np.empty([1, 3])
# Init values
X[0,0] = 70e5
X[0,1] = 30e5
X[0,2] = 0.01

x0 = vertcat(X[0,0], X[0,1], X[0,2])

# sim_model is a class in fp_lib
esp_system = sim_model(rhs, x0)


#these lines of code create the identificaton signal:
simtime = 5 # minutes
T = 1/30
 # Samples pr minute
sim_length = int(simtime/T)


some_input = np.full((sim_length,  2), (0.7484, 61))
# some_input[120:240] = np.full((120, 2), (0.6,52))
# some_input[240:360] = np.full((120, 2), (0.6,54))
# some_input[360:480] = np.full((120, 2), (0.6,56))
# some_input[480:600] = np.full((120, 2), (0.6,58))
# some_input[600:720] = np.full((120, 2), (0.6,60))
# some_input[120:240] = np.full((120, 2), (0.6,60))
# some_input[120:240] = np.full((120, 2), (0.6,60))
# #some_input[60:] = np.full((1140, 2), (0.6,60))
    

#now we need to run the system with that signal to obtain the system response, which is recorded in y_data
y_data = np.empty([sim_length,3])

#the simulation and plotting for data gathering is as follows:
for i in range(sim_length):
    y = esp_system.model_output(some_input[i]).T
    #print(y)
    y_data[[i],] = y

np.save('some_data.npy', some_input) # for testing purposes, comparing it with other framworks.


time = np.linspace(0,simtime,len(y_data)-1)

# Plots the system with the random inputs. Just too see what dataset we are working with

plt.figure(6)
plt.subplot(511)
plt.plot(time, y_data[1:,0]/10**5,color='blue', label='$p_{bh}$')
plt.grid()
plt.legend()
plt.ylabel('$p_{bh} \ [bar]$')


plt.subplot(512)
plt.plot(time, y_data[1:,1]/10**5,color='blue',label='$p_{wh}$')
plt.grid()
plt.legend()
plt.ylabel('$p_{wh} \ [bar]$')

plt.subplot(513)
plt.plot(time, y_data[1:,2]*3600,color='blue',label='$q$')
plt.grid()
plt.legend()
plt.ylabel(r'$q \ [\dfrac{m^3}{h}]$')

plt.subplot(514)
plt.step(time, some_input[1:,0]*100,color='red',label='$z$')
plt.grid()
plt.legend()
plt.ylabel(r'$z \ [\%]$')

plt.subplot(515)
plt.step(time, some_input[1:,1] ,color='red',label='$f$')
plt.grid()
plt.legend()
plt.ylabel(r'$freq \ [Hz]$')
plt.xlabel('Time [min]')


# plt.figure(3)
# plt.rcParams.update({'font.size': 14})
# #plt.subplot(511)
# plt.plot(time, y_data[0:sim_length-1,1]/10**5,color='blue', label='$p_{wh}$')
# plt.grid()
# plt.legend()
# plt.ylabel('$p_{wh} \ [bar]$')
# plt.xlabel('Time [min]')

