# -*- coding: utf-8 -*-
"""
Created on Wed Jun 10 13:16:14 2020

@author: iver_

This program is designed to search for ESN parameters. It compares the ESN with the plant and gives and error.
"""

import numpy as np
import matplotlib.pyplot as plt
import numpy.linalg as la
from control import *
from casadi import *
from fp_lib import *
import scipy.io as io

class sim_model(SimulationModel):
    
    def __init__(self, rhs, x0):
        
        self.rhs = rhs
        
        self.x = x0
        
        
    def model_output(self, u):
        y = f(self.x, u)
        self.x = self.x + y*T
        if self.x[2] < 0:
            self.x[2] = 1e-9
        #print(f)
        return self.x
        
g = 9.81
Cc = 2*10**(-5)
A1 = 0.008107
A2 = 0.008107
D1 = 0.1016
D2 = 0.1016
h1 = 200
h2 = 800
hw = 1000
L1 = 500
L2 = 1200
V1 = 4.054
V2 = 9.729
#ESP DATA
B1 = 1.5*10**9
B2 = 1.5*10**9
M = 1.992*10**8
rho = 950
pr = 1.26 * 10 ** 7
f0 = 60 # Hz
#f = [57, 57, 54, 64, 64, 49, 49, 51, 53, 54]
#f = 53 # Hz
#UNKNOWN
PI = 2.32*10**(-9)
#KANSKJE
pm = 20e5
my = 0.025 # Pa*s
CQ = 1 - 2.6266*my + 6.0032*my**2 - 6.8104*my**3 + 2.7944*my**4
CH = 1 - 0.03*my



x = MX.sym('x',3)
p = MX.sym('p',2)
n_states = x.size1()
n_controls = p.size1()


rhs = vertcat((B1/V1)*(PI*(pr-x[0])-x[2]),(B2/V2)*(x[2]-Cc*(sqrt(fmax(10**-5, x[1]-pm))*p[0])),(1/M)*(x[0]-x[1]-rho*g*hw - 0.158*((rho*L1*x[2]**2)/(D1*A1**2))*(my/(rho*D1*x[2]))**(1/4) - 0.158*((rho*L2*x[2]**2)/(D2*A2**2))*(my/(rho*D2*x[2]))**(1/4) + rho*g* (CH*(9.5970e2 + 7.4959e3*((x[2]/CQ)*(f0/p[1])) - 1.2454e6*((x[2]/CQ)*(f0/p[1]))**2)*(p[1]/f0)**2)))

f = Function('f', [x,p], [rhs])

X = np.empty([1, 3])
# Init values
X[0,0] = 70e5
X[0,1] = 30e5
X[0,2] = 0.01

x0 = vertcat(X[0,0], X[0,1], X[0,2])

esp_system = sim_model(rhs, x0)


#these lines of code create the identificaton signal:
simtime = 2000 # minutes
T = 1/12 # Samples pr minute
sim_length = int(simtime/T)

u_min_valve = np.array([0.0])
u_max_valve = np.array([1.0])
some_input_valve_quick  = RFRAS(u_min_valve,u_max_valve,int(sim_length/2),2)
some_input_valve_slow = RFRAS(u_min_valve, u_max_valve, int(sim_length/2),2)
some_input_valve = np.hstack((some_input_valve_quick, some_input_valve_slow))


u_min_freq = np.array([35])
u_max_freq = np.array([65])
some_input_freq_quick = RFRAS(u_min_freq,u_max_freq,int(sim_length/2),2)
some_input_freq_slow = RFRAS(u_min_freq,u_max_freq,int(sim_length/2),2)
some_input_freq = np.hstack((some_input_freq_quick, some_input_freq_slow))


#THe data MUST be 2dimensional, and data must organized in row form, so in this case:
some_input = np.vstack((some_input_valve, some_input_freq)).T


#the resulting dimension should always be (number of data x number of features)

#now we need to run the system with that signal to obtain the system response, which is recorded in y_data
y_data = np.empty([sim_length,3])

#the simulation and plotting for data gathering is as follows:
for i in range(sim_length):
    y = esp_system.model_output(some_input[i]).T
    #print(y)
    y_data[[i],] = y

#np.save('some_data.npy', some_input) # for testing purposes, comparing it with other framworks.


# time = np.linspace(0,simtime,len(y_data)-1)

# # Plots the system with the random inputs. Just too see what dataset we are working with

# plt.figure(4)
# plt.subplot(511)
# plt.plot(time, y_data[1:,0]/10**5,color='blue', label='$p_{bh}$')
# plt.grid()
# plt.legend()
# plt.ylabel('$p_{bh} \ [bar]$')


# plt.subplot(512)
# plt.plot(time, y_data[1:,1]/10**5,color='blue',label='$p_{wh}$')
# plt.grid()
# plt.legend()
# plt.ylabel('$p_{wh} \ [bar]$')

# plt.subplot(513)
# plt.plot(time, y_data[1:,2]*3600,color='blue',label='$q$')
# plt.grid()
# plt.legend()
# plt.ylabel(r'$q \ [\dfrac{m^3}{h}]$')

# plt.subplot(514)
# plt.step(time, some_input[1:,0]*100,color='red',label='$z$')
# plt.grid()
# plt.legend()
# plt.ylabel(r'$z$ [%]')

# plt.subplot(515)
# plt.step(time, some_input[1:,1] ,color='red',label='$f$')
# plt.grid()
# plt.legend()
# plt.ylabel(r'$f$ [Hz]')
# plt.xlabel('Time [min]')
########################################################
# Normalize
u_max = np.vstack((u_max_valve, u_max_freq)).T
u_min = np.vstack((u_min_valve, u_min_freq)).T
u_data_normal = feature_scaling(some_input,u_max,u_min)

y_max = np.max(y_data,axis=0)
y_min = np.min(y_data,axis=0)
y_data_normal = feature_scaling(y_data,y_max,y_min)

# Creating echo state network
esp_esn = EchoStateNetwork(neu = 200, n_in = 2, n_out = 3, gama=0.12,
                 ro=0.99, psi=0.0,in_scale=0.1, bias_scale=0.1,
                 output_feedback = False, noise_amplitude = 0,
                 load_initial = False, save_initial=False)




regularization = 1e-8
warmupdrop = 50
# regularization = 1e-8
# esp_esn.offline_training1(u_data_normal,y_data_normal,regularization,warmupdrop)
# esp_esn.reset()
# Train network with training data
#esp_esn.offline_training1(u_data_normal,y_data_normal,regularization,warmupdrop)

#esp_esn.reset() # RESET WILL DESTROY THE START, DON'T USE IT!!

esp_esn.add_data(u_data_normal,y_data_normal,warmupdrop)


# #now, we train. The training function also returns the best regularization parameter while insternally changing the weights.
min_reg = 1e-8
max_reg = 1e-9
error, reg = esp_esn.cum_train_cv(min_reg,max_reg)

print("best cv error:",error,"reg value:",reg)




#adter training the ESN, it is recommended you generate another input signal to test the ESN:

test_points = 400

# Creating random inputs
test_u_min_valve = np.array([0.1])
test_u_max_valve = np.array([1.0])
test_input_valve = RFRAS(test_u_min_valve,test_u_max_valve,test_points,5)

test_u_min_freq = np.array([35])
test_u_max_freq = np.array([65])
test_input_freq = RFRAS(test_u_min_freq,test_u_max_freq,test_points,5)


#THe data MUST be 2dimensional, and data must organized in row form, so in this case:
test_input = np.vstack((test_input_valve, test_input_freq)).T

y_pred_test = np.empty([test_points,3])
y_test = np.empty_like(y_pred_test)
#this is where the function feature_descaling comes in handing, as it descales the output back into system scale.


esp_esn.reset()
###################################################
#These lines below could work as a warmup for the prediction. (DO not use!)
modwarmup = 500
warmup_input = np.full((modwarmup,  2), (0.5, 40))

for i in range(0,modwarmup):
    k = esp_system.model_output(warmup_input[i]).T
    u_normal = feature_scaling(warmup_input[i],u_max,u_min)
    esp_esn.update(u_normal)

for i in range(test_points):
    y = esp_system.model_output(test_input[i]).T
    y_test[[i],] = y
    #normalize to apply in ESN
    u_normal = feature_scaling(test_input[i],u_max,u_min)
    y_normal = esp_esn.update(u_normal)
    y_pred_test[i] = feature_descaling(y_normal.flatten(),y_max,y_min)
    

plt.rcParams.update({'font.size': 14})
   
time = np.linspace(0,test_points*T,len(y_test))
#time_test = time = np.linspace(0,20,240)

plt.figure(1)
plt.subplot(311)
# plt.plot(y_test[200:,0]/10**5,color='blue', label='$p_{bh}$ real')
plt.plot(time, y_test[0:,0]/10**5,color='blue')#, label='$p_{bh}$ real')
plt.plot(time, y_pred_test[0:,0]/10**5,color='red')#, label='$p_{bh}$ pred')
plt.grid()
plt.legend()
plt.ylabel('$p_{bh} \ [bar]$')


plt.subplot(312)
# plt.plot(y_test[200:,1]/10**5,color='blue',label='$p_{wh}$ real')
plt.plot(time,y_test[0:,1]/10**5,color='blue')#,label='$p_{wh}$ real')
plt.plot(time,y_pred_test[0:,1]/10**5,color='red')#,label='$p_{wh}$ pred')
plt.grid()
plt.legend()
plt.ylabel('$p_{wh} \ [bar]$')

plt.subplot(313)
# plt.plot(y_test[200:,2]*3600,color='blue',label='$q$ real')
plt.plot(time,y_test[0:,2]*3600 ,color='blue')#,label='$q$ real')
plt.plot(time,y_pred_test[0:,2]*3600,color='red')#,label='$q$ pred')
plt.grid()
plt.legend()
plt.xlabel(r'Time [Min]')
plt.ylabel(r'$q \ [\dfrac{m^3}{h}]$')


# Not a good idea with mean square error because of the huge diffrenece in values.
# error = abs((y_test - y_pred_test))**2
# mean_error = (1/test_points)*error 

# print("NRMSE: ", mean_error)

y_normal_error = feature_scaling(y_test, y_max, y_min)
y_pred_normal_error = feature_scaling(y_pred_test, y_max, y_min)

# Relative Prior Error (persentage)

# y_error = np.zeros((len(y_test)))
# for k in range(test_points):
#     y_error[k] = 100 *  (abs(y_normal_error[k,0]-y_pred_normal_error[k,0])/(y_normal_error[k,0]))
    

y_error = np.zeros((len(y_test)))
for k in range(test_points):
    y_error[k] = 100 * (abs(y_test[k,0]-y_pred_test[k,0])/(y_test[k,0]))


plt.figure(2)
# plt.plot(y_test[200:,0]/10**5,color='blue', label='$p_{bh}$ real')
plt.plot(time, y_error,color='red')
plt.grid()
plt.legend()
plt.ylabel('Error $[\%]$')
plt.xlabel(r'Time [Min]')


# # Create data set for simulation results.
# esnstates = np.zeros((length, 3))
# esnstates = y_pred

# plantstates = np.zeros((length, 3))
# plantstates = xx.T

# inputs = np.zeros((length, 2))
# inputs = u_cl

# cost_function = np.zeros((length))
# cost_function = f_val


io.savemat('errorfastFAST.mat', {'plant': y_test, 'esn': y_pred_test, 'input': test_input, 'error': y_error})



# error = 0
# for k in range(test_points):
#     error = error + abs(y_normal_error[k,0]-y_pred_normal_error[k,0]) + abs(y_normal_error[k,0]-y_pred_normal_error[k,0])  + abs(y_normal_error[k,2]-y_pred_normal_error[k,2]) 

# print("Error: ", error)

# # Reservoir size errors
# node = [20, 50, 100, 200, 400]
# error = [121, 91, 50, 37, 33]
# plt.title('Error plot')
# plt.errorbar(node, error,fmt='o',color='red')
# plt.ylabel(r'Sum of normalized error')
# plt.xlabel(r'Size of reservoir (Neurons)')


# # Leak rate errors
# leak_rate = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1]
# error = [44.6, 36.0, 48.2, 74.4, 81.7, 89.8, 96.4, 96.9, 122.9, 161.2]
# plt.title('Leak rate error plot')
# plt.errorbar(leak_rate, error, color='red')
# plt.ylabel(r'Sum of normalized error')
# plt.xlabel(r'Leak rate $\alpha$')

# # Leak rate error zoomed
# leak_rate = [0.1, 0.12, 0.14, 0.16, 0.18, 0.2, 0.22, 0.24, 0.26, 0.28, 0.3]
# error = [44.6, 33.5, 37.8, 39.1, 36.4, 36, 39.4, 40.2, 44.9, 44.9, 48.2]
# plt.title('Leak rate error plot')
# plt.errorbar(leak_rate, error,color='red')
# plt.ylabel(r'Sum of normalized error')
# plt.xlabel(r'Leak rate $\alpha$')